@@ -3,7 +3,7 @@
 
 """Unittests for the posix1e module"""
 
-#  Copyright (C) 2002-2009, 2012 Iustin Pop <iusty@k1024.org>
+#  Copyright (C) 2002-2009, 2012, 2014, 2015 Iustin Pop <iustin@k1024.org>
 #
 #  This library is free software; you can redistribute it and/or
 #  modify it under the terms of the GNU Lesser General Public
@@ -26,6 +26,8 @@ import os
 import tempfile
 import sys
 import platform
+import re
+import errno
 
 import posix1e
 from posix1e import *
@@ -40,6 +42,9 @@ M0500 = 320 # octal 0500
 M0644 = 420 # octal 0644
 M0755 = 493 # octal 755
 
+# Check if running under Python 3
+IS_PY_3K = sys.hexversion >= 0x03000000
+
 def _skip_test(fn):
     """Wrapper to skip a test"""
     new_fn = lambda x: None
@@ -54,6 +59,20 @@ def has_ext(extension):
     else:
         return lambda x: x
 
+def ignore_ioerror(errnum, fn, *args, **kwargs):
+    """Call a function while ignoring some IOErrors.
+
+    This is needed as some OSes (e.g. FreeBSD) return failure (EINVAL)
+    when doing certain operations on an invalid ACL.
+
+    """
+    try:
+        fn(*args, **kwargs)
+    except IOError:
+        err = sys.exc_info()[1]
+        if err.errno == errnum:
+            return
+        raise
 
 class aclTest:
     """Support functions ACLs"""
@@ -210,7 +229,7 @@ class ModificationTests(aclTest, unittest.TestCase):
 
     def testStr(self):
         """Test str() of an ACL."""
-        acl = posix1e.ACL()
+        acl = posix1e.ACL(text=BASIC_ACL_TEXT)
         str_acl = str(acl)
         self.checkRef(str_acl)
 
@@ -220,7 +239,7 @@ class ModificationTests(aclTest, unittest.TestCase):
         acl = posix1e.ACL()
         e = acl.append()
         e.tag_type = posix1e.ACL_OTHER
-        acl.calc_mask()
+        ignore_ioerror(errno.EINVAL, acl.calc_mask)
         str_format = str(e)
         self.checkRef(str_format)
 
@@ -230,9 +249,9 @@ class ModificationTests(aclTest, unittest.TestCase):
         acl = posix1e.ACL()
         e = acl.append()
         e.tag_type = posix1e.ACL_OTHER
-        acl.calc_mask()
+        ignore_ioerror(errno.EINVAL, acl.calc_mask)
         acl.delete_entry(e)
-        acl.calc_mask()
+        ignore_ioerror(errno.EINVAL, acl.calc_mask)
 
     @has_ext(HAS_ACL_ENTRY)
     def testDoubleEntries(self):
@@ -269,11 +288,11 @@ class ModificationTests(aclTest, unittest.TestCase):
     @has_ext(HAS_ACL_ENTRY)
     def testMultipleBadEntries(self):
         """Test multiple invalid entries"""
-        acl = posix1e.ACL(text=BASIC_ACL_TEXT)
-        self.assertTrue(acl.valid(), "ACL built from standard description"
-                        " should be valid")
         for tag_type in (posix1e.ACL_USER,
                          posix1e.ACL_GROUP):
+            acl = posix1e.ACL(text=BASIC_ACL_TEXT)
+            self.assertTrue(acl.valid(), "ACL built from standard description"
+                                         " should be valid")
             e1 = acl.append()
             e1.tag_type = tag_type
             e1.qualifier = 0
@@ -285,11 +304,13 @@ class ModificationTests(aclTest, unittest.TestCase):
             e2.tag_type = tag_type
             e2.qualifier = 0
             e2.permset.clear()
-            acl.calc_mask()
+            ignore_ioerror(errno.EINVAL, acl.calc_mask)
             self.assertFalse(acl.valid(), "ACL should not validate when"
                 " containing two duplicate entries")
             acl.delete_entry(e1)
-            acl.delete_entry(e2)
+            # FreeBSD trips over itself here and can't delete the
+            # entry, even though it still exists.
+            ignore_ioerror(errno.EINVAL, acl.delete_entry, e2)
 
     @has_ext(HAS_ACL_ENTRY)
     def testPermset(self):
@@ -320,5 +341,56 @@ class ModificationTests(aclTest, unittest.TestCase):
                 " after deletion" % pmap[perm])
 
 
+    @has_ext(HAS_ACL_ENTRY and IS_PY_3K)
+    def testQualifierValues(self):
+        """Tests qualifier correct store/retrieval"""
+        acl = posix1e.ACL()
+        e = acl.append()
+        # work around deprecation warnings
+        if hasattr(self, 'assertRegex'):
+            fn = self.assertRegex
+        else:
+            fn = self.assertRegexpMatches
+        for tag in [posix1e.ACL_USER, posix1e.ACL_GROUP]:
+            qualifier = 1
+            e.tag_type = tag
+            while True:
+                if tag == posix1e.ACL_USER:
+                    regex = re.compile("user with uid %d" % qualifier)
+                else:
+                    regex = re.compile("group with gid %d" % qualifier)
+                try:
+                    e.qualifier = qualifier
+                except OverflowError:
+                    # reached overflow condition, break
+                    break
+                self.assertEqual(e.qualifier, qualifier)
+                fn(str(e), regex)
+                qualifier *= 2
+
+    @has_ext(HAS_ACL_ENTRY and IS_PY_3K)
+    def testQualifierOverflow(self):
+        """Tests qualifier overflow handling"""
+        acl = posix1e.ACL()
+        e = acl.append()
+        qualifier = sys.maxsize * 2
+        for tag in [posix1e.ACL_USER, posix1e.ACL_GROUP]:
+            e.tag_type = tag
+            with self.assertRaises(OverflowError):
+                e.qualifier = qualifier
+
+    @has_ext(HAS_ACL_ENTRY and IS_PY_3K)
+    def testNegativeQualifier(self):
+        """Tests negative qualifier handling"""
+        # Note: this presumes that uid_t/gid_t in C are unsigned...
+        acl = posix1e.ACL()
+        e = acl.append()
+        for tag in [posix1e.ACL_USER, posix1e.ACL_GROUP]:
+            e.tag_type = tag
+            for qualifier in [-10, -5, -1]:
+                with self.assertRaises(OverflowError):
+                    e.qualifier = qualifier
+
+
 if __name__ == "__main__":
     unittest.main()
